# -*- coding: utf-8 -*-
"""
Provide mathematical functions set up to process data from Audio objects.

Created on Tue May  5 00:34:36 2020

@author: João Vitor Gutkoski Paes
"""

import numpy as np
import numba as nb


@nb.njit(parallel=True)
def apply_channelwise(auddata: np.ndarray, func: callable) -> np.ndarray:
    """
    Apply func on each column of an audio data array.

    Parameters
    ----------
    auddata : np.ndarray
        Audio data as numpy array.
    func : callable
        The function to apply on each channel. Must receive data as a 1d array.

    Returns
    -------
    chwise : np.ndarray
        Output of func for each channel as an array.

    """
    chwise = np.zeros((1, auddata.shape[1]), dtype=auddata.dtype)
    for channel in nb.prange(chwise.shape[1]):
        chwise[:, channel] = func(auddata[:, channel])
    return chwise



@nb.njit
def max_abs(auddata: np.ndarray) -> np.ndarray:
    """
    Maximum of the absolute values of the input array, for each channel (column).

    See `apply_channelwise` for more info

    Parameters
    ----------
    auddata : np.ndarray
        Input array of audio data.

    Returns
    -------
    ma : np.ndarray
        Columnwise `max` of `abs` of `arr`.

    """
    ma = apply_channelwise(auddata, lambda arr: np.max(np.abs(arr)))
    return ma


@nb.njit
def rms(auddata: np.ndarray) -> np.ndarray:
    """
    Root of the mean of the squared values of input array, for each channel.

    See `apply_channelwise` for more info.

    Parameters
    ----------
    auddata : np.ndarray
        Input array of audio data.

    Returns
    -------
    rms : np.ndarray
        Columnwise `sqrt` of `mean` of `arr**2`.

    """
    rms = apply_channelwise(auddata, lambda arr: np.mean(arr**2)**0.5)
    return rms


@nb.njit
def dB(auddata: np.ndarray, power: bool = False, ref: float = 1.0) -> np.ndarray:
    """
    Decibel level of input array, for each channel.

    See `apply_channelwise` and `rms` for more info.

    Parameters
    ----------
    auddata : np.ndarray
        Input array of audio data.
    power : bool, optional
        If True, assumes a power signal, which need not to be squared. The default is False.
    ref : float, optional
        The decibel reference. The default is 1.0.

    Returns
    -------
    _np.ndarray
        Channelwise audio levels, in decibel.

    """
    return (10 if power else 20) * np.log10(auddata / ref)


@nb.njit
def noise(level: float, samplerate: int, tlen: float, nchannels: int) -> np.ndarray:
    """
    Generate random noise for audio data.

    Parameters
    ----------
    level : float
        Level of audio.
    samplerate : int
        Sample rate.
    tlen : float
        Total time length.
    nchannels : int
        Total number of channels.

    Returns
    -------
    _np.ndarray
        The complete noise data.

    """
    shape = (int(samplerate*tlen), nchannels)
    noise = np.zeros(shape, dtype=np.float32)
    noise[:] = np.random.randn(*shape)
    noise[:, ] /= max_abs(noise)
    return level*noise    # TODO: DEVELOP FOR DECIBEL GAIN LEVEL.




# -*- coding: utf-8 -*-
"""
This module provide two classes: `Now` and `Logger`.

The first is just a wrap around time.localtime() with human readable interface of the exact time instant the object is created.
The latter is a simple file logger that uses `Now` instances to record information.

Created on Sat Jun  6 00:39:46 2020

@author: joaovitor
"""

# import json
import time
import random as rd
from typing import TextIO


class Now(object):
    """Human readable data about date and time."""
    _keys = ('year', 'month', 'day', 'wday', 'hour', 'min', 'sec')
    _wdays = {0: 'mon', 1: 'tue', 2: 'wed', 3: 'thu', 4: 'fri', 5: 'sat', 6: 'sun'}

    def __init__(self):
        localtime = time.localtime()
        self._hour = f"0{localtime.tm_hour}" if localtime.tm_hour < 10 else f"{localtime.tm_hour}"
        self._min = f"0{localtime.tm_min}" if localtime.tm_min < 10 else f"{localtime.tm_min}"
        self._sec = f"0{localtime.tm_sec}" if localtime.tm_sec < 10 else f"{localtime.tm_sec}"
        self._day = f"0{localtime.tm_mday}" if localtime.tm_mday < 10 else f"{localtime.tm_mday}"
        self._month = f"0{localtime.tm_mon}" if localtime.tm_mon < 10 else f"{localtime.tm_mon}"
        self._year = f"{localtime.tm_year}"
        self._wday = f"{Now._wdays[localtime.tm_wday]}"
        self._GMT = f"GMT_{localtime.tm_zone}"
        return

    def __str__(self):
        return f"{self.wday}_{self.date}_{self.time}_{self.GMT}"

    @property
    def date(self):
        return f"{self.day}-{self.month}-{self.year}"

    @property
    def time(self):
        return f"{self.hour}-{self.min}-{self.sec}"

    @property
    def hour(self):
        """Wall clock hours."""
        return self._hour

    @property
    def min(self):
        """Wall clock minutes."""
        return self._min

    @property
    def sec(self):
        """Wall clock seconds."""
        return self._sec

    @property
    def year(self):
        """Year."""
        return self._year

    @property
    def month(self):
        """Month."""
        return self._month

    @property
    def day(self):
        """Month day."""
        return self._day

    @property
    def wday(self):
        """Week day."""
        return self._wday

    @property
    def GMT(self):
        """Time zone."""
        return self._GMT


class Logger(object):
    """Simple logger."""

    _randlogs = ['Albatroz', 'Jaguatirica', 'Jibóia', 'Arara', 'Mico', 'Boitatá', 'Boto Cor-de-Rosa']

    def __init__(self, name: str = 'common', ext: str = 'log',
                 title: str = 'title', logend: str = 'end.') -> None:
        """
        File based logger.

        Parameters
        ----------
        name : str, optional
            File name. The default is 'common'.
        ext : str, optional
            File extension. The default is 'log'.
        title : str, optional
            A title or description for the log. The default is 'title'.
        logend : str, optional
            A tag that represents the end of a logging session.
            The default is 'end.'.

        Returns
        -------
        None

        """
        self._name = name
        self._ext = ext
        self._title = title
        self._logend = f"\n# {logend}\n{13*'-'}\n\n"
        self._time = Now()
        self._header = f'{13*"-"}\n# LOG OUTPUT FILE\n# {self.name}\n# {self.time}\n\n# {self.title}\n'
        self.fopen()
        return

    def __del__(self):
        """Close underlying file on del statement."""
        try:
            self.fclose()
        except AttributeError:
            pass
        return

    @property
    def name(self) -> str:
        """Log file name."""
        return self._name

    @property
    def ext(self) -> str:
        """File extension."""
        return self._ext

    @property
    def title(self) -> str:
        """Log title."""
        return self._title

    @property
    def time(self) -> str:
        """Time at log creation."""
        return str(self._time)

    @property
    def header(self) -> str:
        """File header."""
        return self._header

    @property
    def logend(self) -> str:
        """End of logging."""
        return self._logend

    @property
    def file(self) -> TextIO:
        """File pointer."""
        return self._file

    def fopen(self):
        """Open the log file, enabling `read` and `write`."""
        self._file = open(f'{self.name}.{self.ext}', mode='a+')
        return

    def start_log(self):
        """Write header to log file."""
        self.file.write(self.header)
        return

    def log(self, message: str = None) -> int:
        """
        Write `message` to the log file along with the current time.

        Parameters
        ----------
        message : str, optional
            The content of the logging. If no input is given, randomizes an item from `LogFile._randlog`.
            The default is None.

        Returns
        -------
        b : int
            The amount of bytes written to log file.

        """
        time = Now().time
        msg = message if not message is None \
            else Logger._randlogs[rd.randint(0, len(Logger._randlogs)-1)]
        return self.file.write(f"{time} : {msg}\n")

    def end_log(self):
        """Write the end of logging tag."""
        self.file.write(self.logend)
        return

    def fclose(self):
        """Close the log file."""
        self.file.close()
        del self._file
        return

    def print_log(self):
        """Print the log file content. File must be opened."""
        try:
            self.file.seek(0)
            print(self.file.read())
        except AttributeError:
            self.fopen()
            self.file.seek(0)
            print(self.file.read())
            self.fclose()
        return


def _random_log(logger, timeout):
    init = time.time()
    while (init + timeout) > time.time():
        time.sleep(0.125)
        logger.log()
    logger.end_log()
    return


if __name__ == '__main__':
    NAME = 'log'
    EXT = 'txt'
    FILENAME = f'{NAME}.{EXT}'
    TITLE = 'Overall Level'

    logger = Logger(NAME, EXT, TITLE)
    _random_log(logger, 2.)
    logger.fclose()

    logger.reopen()
    _random_log(logger, 3.)
    logger.fclose()



# -*- coding: utf-8 -*-
"""
Simple module to color texts on command line output.

@author: João Vitor Gutkoski Paes
"""



__STRCLR = {
    "none": "",
    "black": "\x1b[30m",
    "red": "\x1b[31m",
    "green": "\x1b[32m",
    "yellow": "\x1b[33m",
    "blue": "\x1b[34m",
    "pink": "\x1b[35m",
    "cyan": "\x1b[36m",
    "white": "\x1b[37m",
    "clear": "\x1b[0m"
}


__BGRCLR = {
    "none": "",
    "black": "\x1b[40m",
    "red": "\x1b[41m",
    "green": "\x1b[42m",
    "yellow": "\x1b[43m",
    "blue": "\x1b[44m",
    "pink": "\x1b[45m",
    "cyan": "\x1b[46m",
    "white": "\x1b[47m",
    "clear": "\x1b[0m"
}


_LINEOBJ = {
    "clear": "0",
    "font": "3",
    "background": "4",
}


_COLORS = {
    "clear": "",
    "black": "0",
    "red": "1",
    "green": "2",
    "yellow": "3",
    "blue": "4",
    "pink": "5",
    "cyan": "6",
    "white": "7"
}


class Error(Exception):
    pass


class ColorNameError(Error):
    def __init__(self):
        """
        Error raised if color not implemented.

        Returns:
            None.

        """
        self.message = "This color is not available."
        return


class ColorStr(object):
    """Color string"""
    
    def __init__(self, font: str = "clear", back: str = "clear") -> None:
        """
        Pintor de linhas.

            Example use:

                >>> black_on_white_str = ColorStr("black", "white")
                >>> red_on_green_str = ColorStr("red", "green")
                >>> print(black_on_white_str("Estou usando colore.ColorStr"))
                >>> print(red_on_green_str("I'm using colore.ColorStr"))

        Args:
            font (str, optional): Cor da fonte | Font color. Defaults to "clear".
            back (str, optional): Cor do fundo | Background color. Defaults to "clear".

        Returns:
            None.

        """
        self.fntclr = font
        self.bgrclr = back
        return

    def __call__(self, text: str = None) -> str:
        """
        Paint the text with its font and backgroud colors.

        Args:
            text (str, optional): The text to be painted. Defaults to None.

        Returns:
            str: Colored text and background.

        """
        if text is None:
            text = ColorStr.__str__(self)
        return colorir(text, self.fntclr, self.bgrclr)

    @property
    def fntclr(self) -> str:
        """Font color."""
        return self._fntclr

    @fntclr.setter
    def fntclr(self, color: str) -> None or Error:
        if color in _COLORS.keys():
            self._fntclr = color
        else:
            raise ColorNameError

    @property
    def font(self) -> str:
        """Alias for fntclr."""
        return self.fntclr

    @property
    def bgrclr(self) -> str:
        """Background color."""
        return self._bgrclr

    @bgrclr.setter
    def bgrclr(self, color: str) -> None or Error:
        if color in _COLORS.keys():
            self._bgrclr = color
        else:
            raise ColorNameError

    @property
    def background(self) -> str:
        """Alias for bgrclr."""
        return self.bgrclr

    @property
    def back(self) -> str:
        """Alias for bgrclr."""
        return self.bgrclr


def _color_code(which: str, color: str) -> str:
    return f'\x1b[{_LINEOBJ[which]}{_COLORS[color]}m'


def _color_clear() -> str:
    return _color_code("clear", "clear")


def _pintor(text: str, which: str, color: str) -> str or Error:
    try:
        return f'{_color_code(which, color)}{text}{_color_clear()}'
    except KeyError:
        raise ColorNameError


def pinta_texto(text: str, color: str) -> str:
    """
    Pinta o texto com a cor escolhida.

    Args:
        text (str): O texto em si.
        color (str): A cor escolhida.

    Returns:
        str: Texto colorido.

    """
    return _pintor(text, "font", color)


def pinta_fundo(text: str, color: str) -> str:
    """
    Pinta o fundo do texto com a cor escolhida.

    Args:
        text (str): O texto em si.
        color (str): A cor escolhida.

    Returns:
        str: Texto com fundo colorido.

    """
    return _pintor(text, "background", color)


def colorir(texto: str, fntclr: str = None, bgrclr: str = None) -> str:
    """
    Retorna o texto colorido nas cores escolhidas.

    Caso nenhuma cor seja informada, retorna o texto sem alterações.

    Args:
        texto (str): DESCRIPTION.
        fntclr (str, optional): Font color. Defaults to None.
        bgrclr (str, optional): Background color. Defaults to None.

    Returns:
        texto (str): Colored font and background.

    """
    if fntclr is None:
        fntclr = "clear"
    elif bgrclr is None:
        bgrclr = "clear"
    texto = pinta_texto(texto, fntclr)
    texto = pinta_fundo(texto, bgrclr)
    return texto





# -*- coding: utf-8 -*-
"""
Audio IO with easy access to memory buffers for real-time analysis and processing of data.

Author: João Vitor Gutkoski Paes
"""

from . import utils
from .configurations import Configurations
from .audio import Audio, AudioBuffer
from .streamer import Recorder, Player
from .monitor import Monitor

__all__ = ['Audio', 'AudioBuffer',
           'Recorder', 'Player',
           'Monitor',
           'Configurations',
           'utils']




# -*- coding: utf-8 -*-
"""
Provides frequency and fractional octave frequency bands functionalities.

@author: João Vitor Gutkoski Paes
"""

import numpy as np
from typing import Tuple

# Cálculo de bandas de oitava a partir de 1 kHz utilizando base 10 ou 2
__nominal_frequencies = np.array([
    0.1, 0.125, 0.16, 0.2, 0.25, 0.315, 0.4, 0.5, 0.6, 3., 0.8,
    1., 1.25, 1.6, 2., 2.5, 3.15, 4., 5., 6.3, 8., 10., 12.5, 16.,
    20., 25., 31.5, 40., 50., 63., 80., 100., 125., 160., 200., 250.,
    315., 400., 500., 630., 800., 1000., 1250., 1600., 2000., 2500.,
    3150., 4000., 5000., 6300., 8000., 10000., 12500., 16000., 20000.
])


def freq_to_band(freq: float, nthOct: int, ref: float, base: int) -> int:
    """
    Band number from frequency value.

    Parameters
    ----------
    freq : float
        The frequency value.
    nthOct : int
        How many bands per octave.
    ref : float
        Frequency of reference, or band number 0.
    base : int
        Either 10 or 2.

    Raises
    ------
    ValueError
        If base is not 10 nor 2 raises value error.

    Returns
    -------
    int
        The band number from center.

    """
    if base == 10:
        log = np.log10
        factor = 3 / 10
    elif base == 2:
        log = np.log2
        factor = 1
    else:
        raise ValueError(f"freq_to_band: unknown base value: {base}.")
    return int(np.round(log(freq / ref) * (nthOct / factor)))


def fractional_octave_frequencies(nthOct: int = 3,
                                  freqRange: Tuple[float] = (20., 20000.),
                                  refFreq: float = 1000.,
                                  base: int = 10) -> np.ndarray:
    """
    Lower, center and upper frequency values of all bands within range.

    Parameters
    ----------
    nthOct : int, optional
        bands of octave/nthOct. The default is 3.
    freqRange : Tuple[float], optional
        frequency range. These frequencies are inside the lower and higher band, respectively.
        The default is (20., 20000.).
    refFreq : float, optional
        Center frequency of center band. The default is 1000..
    base : int, optional
        Either 10 or 2. The default is 10.

    Returns
    -------
    freqs : numpy.ndarray
        Array with shape (N, 3).

    """
    if base == 10:
        factor = 3 / 10
    elif base == 2:
        factor = 1
    minFreq, maxFreq = freqRange
    minBand = freq_to_band(minFreq, nthOct, refFreq, base)
    maxBand = freq_to_band(maxFreq, nthOct, refFreq, base)
    bands = np.arange(minBand, maxBand + 1)
    freqs = np.zeros((len(bands), 3))
    nthOct = 1 / nthOct
    for k, band in enumerate(bands):
        dummy = refFreq * base ** (band * nthOct * factor)
        dummy = np.sqrt((__nominal_frequencies - dummy) ** 2)
        center = __nominal_frequencies[np.argmin(dummy)]
        lower = center / base ** (nthOct * factor / 2)
        upper = center * base ** (nthOct * factor / 2)
        freqs[k, :] = [lower, center, upper]
    return freqs


def normalize_frequencies(freqs: np.ndarray,
                          samplingRate: int = 44100) -> np.ndarray:
    """
    Normalize frequencies for any sampling rate.

    Parameters
    ----------
    freqs : np.ndarray
        DESCRIPTION.
    samplingRate : int, optional
        DESCRIPTION. The default is 44100.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """
    nyq = samplingRate // 2
    return freqs / nyq


def freqs_to_center_and_edges(freqs: np.ndarray) -> Tuple[np.ndarray]:
    """
    Separate the array returned from `fractional_octave_frequencies`.

    The returned arrays corresponde to the center and edge frequencies of
    the fractional octave bands

    Parameters
    ----------
    freqs : np.ndarray
        Array returned from `fractional_octave_frequencies`.

    Returns
    -------
    center : np.ndarray
        Center frequencies of the bands.
    edges : np.ndarray
        Edge frequencies (lower and upper) of the bands.

    """
    center = freqs[:, 1].T
    edges = np.array([freqs[:, 0], freqs[:, 2]]).T
    return center, edges



# -*- coding: utf-8 -*-
"""
Created on Fri May 29 16:08:22 2020

@author: João Vitor Gutkoski Paes
"""

import numpy as np
import wave as wv
from audiodata import AudioData


class WaveFile(object):
    """Wave file object."""

    def __init__(self, filename: str) -> None:
        self.name = filename
        return

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_type:
            raise exc_value
        return

    def write(self, ad: AudioData) -> None:
        with wv.open(self.name, "wb") as wfile:
            wfile.setnchannels(ad.channels)
            wfile.setframerate(ad.samplerate)
            wfile.setnframes(ad.numsamples)
            wfile.setsampwidth(ad.samplesize)
            wfile.writeframes(np.float32(ad.data))
        return

    def read(self) -> AudioData:
        with wv.open(self.name, "rb") as rfile:
            numsamples = rfile.getnframes()
            samplerate = rfile.getframerate()
            databuff = rfile.readframes(numsamples)
            data = np.frombuffer(databuff, dtype=np.float16)
        ad = AudioData(data, samplerate)
        return ad


if __name__ == "__main__":
    SAMPLE_RATE = 44100  # [Hz] one second of data
    TOTAL_TIME = 2
    TOTAL_SAMPLES = TOTAL_TIME * SAMPLE_RATE
    CHANNELS = 1
    FILE_NAME = 'testfile.wav'

    #timesteps = np.linspace(0, TOTAL_TIME, TOTAL_SAMPLES)
    #sine = np.sin(2*np.pi*250*timesteps)

    noise = np.random.randn(TOTAL_SAMPLES, CHANNELS)
    noise /= np.max(np.abs(noise))

    audiow = AudioData(noise, SAMPLE_RATE)

    with Wave(FILE_NAME) as File:
        File.write(audiow)
        audior = File.read()

    print(audior.data.dtype, audiow.data.dtype)



# -*- coding: utf-8 -*-
"""
This module provide a singleton object to hold configurations such as `samplerate`, `blocksize` and any relevant parameter that is necessary to the main `Recorder` and `Player` classes.

These values are used as default case and are available to any OsSom class that might need them.

Created on Sun Jun 28 21:40:13 2020

@author: João Vitor Gutkoski Paes
"""


import numpy as _np
import warnings
from typing import Dict, List


_default = {
    'samplerate': 48000,
    'blocksize': 512,
    'buffersize': 480000,
    'dtype': _np.dtype('float32'),
    'channels': {'in': [0, 1],
                 'out': [0, 1]}
}


class Configurations:
    """Global OsSom configurations."""

    _samplerate: int = None
    _blocksize: int = None
    _buffersize: int = None
    _dtype: _np.dtype = None
    _channels: Dict[str, List[int]] = None

    _instance = None

    def __init__(self):
        """
        This class is a singleton, which means that any "new" instance is actually the same, pointed to by a new name.

            >>> new_config = ossom.Configurations()
            >>> new_config is ossom.config
            True

        Basically it provides a one time setup interface to convinently create `Recorder`s and `Player`s with only default parameters.

            >>> rec1 = Recorder(samplerate=44100)
            >>> config.samplerate = 44100
            >>> rec2 = Recorder()
            >>> rec1.samplerate == rec2.samplerate
            True

        Returns
        -------
        None.

        """
        return

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.reset()
        return cls._instance

    def __repr__(self):
        s = f"Configurations(\n\tsamplerate={self.samplerate},\n\tblocksize={self.blocksize},\n\tbuffersize={self.buffersize},\n\tdtype={self.dtype},\n\tchannels={self.channels})"
        return s

    @property
    def samplerate(self) -> int:
        """Sample rate, or sampling frequency."""
        return self._samplerate

    @samplerate.setter
    def samplerate(self, new: int):
        if 11025 <= new and new <= 192000:
            new = int(new)
            self._samplerate = new
        else:
            raise ValueError("Sample rate out of acceptable range [11025, 192000].")
        return

    @property
    def blocksize(self) -> int:
        """Size of audio chunk read on calls to `next`."""
        return self._blocksize

    @blocksize.setter
    def blocksize(self, new: int):
        self._blocksize = int(new)
        return

    @property
    def buffersize(self) -> int:
        """Total size of the audio buffer held by `Recorder` or `Player` objects."""
        return self._buffersize

    @buffersize.setter
    def buffersize(self, new: int):
        if new < self.samplerate * 5:
            warnings.warn("Buffer may be too short.", ResourceWarning)
        self._buffersize = int(new)
        return

    @property
    def dtype(self) -> _np.dtype:
        """The audio data type."""
        return self._dtype

    @dtype.setter
    def dtype(self, new: str or _np.dtype):
        new = _np.dtype(new).name
        if new == 'float64':
            new = 'float32'
        self._dtype = _np.dtype(new)
        return

    @property
    def channels(self) -> Dict[str, List[int]]:
        """
        Dictionary containing two lists that represents the device channels to be used.
        Zero indexed.
        """
        return self._channels

    @property
    def inChannels(self) -> List[int]:
        """List of requested device input channels. Zero indexed."""
        return self._channels['in']

    @inChannels.setter
    def inChannels(self, new: List[int]):
        if not isinstance(new, list):
            if not isinstance(new, int):
                raise ValueError("The input channels must be an integer or a list of integers representing the device channels. Zero indexed.")
            new = list(range(int))
        self._channels['in'] = new
        return

    @property
    def outChannels(self) -> List[int]:
        """List of requested device output channels. Zero indexed."""
        return self._channels['out']

    @outChannels.setter
    def outChannels(self, new: List[int]):
        if not isinstance(new, list):
            if not isinstance(new, int):
                raise ValueError("The output channels must be an integer or a list of integers representing the device channels. Zero indexed.")
            new = list(range(int))
        self._channels['in'] = new
        return

    def reset(self):
        """Set all configuration values back to module default."""
        self._samplerate: int = _default['samplerate']
        self._blocksize: int = _default['blocksize']
        self._buffersize: int = _default['buffersize']
        self._dtype: _np.dtype = _default['dtype']
        self._channels: List[int] = _default['channels']
        return

config = Configurations()



# -*- coding: utf-8 -*-
"""
Lacks documentation!

Created on Sun Jun 28 21:12:12 2020

@author: João Vitor Gutkoski Paes
"""


import numpy as _np
import soundcard as _sc
import multiprocessing as _mp
import threading as _td
from ossom import Audio, AudioBuffer, Configurations
from typing import List


config = Configurations()


class _Streamer(AudioBuffer):
    """Base streamer class."""

    def __init__(self,
                 samplerate: int,
                 blocksize: int,
                 channels: List[int],
                 buffersize: int,
                 dtype: _np.dtype):
        AudioBuffer.__init__(self, None, samplerate, buffersize,
                             len(channels), blocksize//2, dtype)
        self.running = _mp.Event()
        self.finished = _mp.Event()
        return

    def get_buffer(self, blocksize: int = None):
        pass

    def _loop_wrapper(self, blocking: bool):
        self.finished.clear()
        self.reset()
        self._thread = _td.Thread(target=self._loop)
        self._thread.start()
        if blocking:
            self.finished.wait()
            self.stop()
        return


class Recorder(_Streamer):
    """Recorder class."""

    def __init__(self, id: int or str = None,
                 samplerate: int = config.samplerate,
                 blocksize: int = config.blocksize,
                 channels: List[int] = config.channels['in'],
                 buffersize: int = config.buffersize,
                 dtype: _np.dtype = config.dtype,
                 loopback: bool = False):
        """
        Record audio from input device directly into shared memory.



        Parameters
        ----------
        id : int or str, optional
            DESCRIPTION. The default is None.
        samplerate : int, optional
            DESCRIPTION. The default is config.samplerate.
        blocksize : int, optional
            DESCRIPTION. The default is config.blocksize.
        channels : List[int], optional
            DESCRIPTION. The default is config.channels.
        buffersize : int, optional
            DESCRIPTION. The default is config.buffersize.
        dtype : _np.dtype, optional
            DESCRIPTION. The default is config.dtype.
        loopback : bool, optional
            DESCRIPTION. The default is False.

        Returns
        -------
        None.

        """
        _Streamer.__init__(self, samplerate, blocksize, channels, buffersize, dtype)
        self._channels = channels
        self._mic = _sc.default_microphone() if not id \
            else _sc.get_microphone(id, include_loopback=loopback)
        return

    def __call__(self, tlen: float = 5., blocking: bool = False):
        self.frames = int(_np.ceil(tlen * self.samplerate))
        if self.frames > self.nsamples:
            raise MemoryError("Requested recording time is greater than available space.")
        self._loop_wrapper(blocking)
        return

    @property
    def channels(self):
        """The device channels to record from. Zero indexed."""
        return self._channels

    def _loop(self):
        with self._mic.recorder(self.samplerate, self.channels, self.blocksize) as r:
            self.running.set()
            while self.widx < self.frames:
                self.write_next(r.record(self.blocksize//4))
                if self.finished.is_set() or self.is_full:
                    break
            r.flush()
        self.running.clear()
        self.finished.set()
        return

    def stop(self):
        if not self.finished.is_set():
            self.finished.set()
        self._thread.join()
        return

    def reset(self):
        self.widx = 0
        return

    def get_record(self, blocksize: int = None):
        return Audio(self.data[:self.frames].copy(), self.samplerate,
                     self.blocksize if not blocksize else blocksize)


class Player(_Streamer):
    def __init__(self, id: int or str = None,
                 samplerate: int = config.samplerate,
                 blocksize: int = config.blocksize,
                 channels: List[int] = config.channels['out'],
                 buffersize: int = config.buffersize,
                 dtype: _np.dtype = config.dtype):
        _Streamer.__init__(self, samplerate, blocksize, channels, buffersize, dtype)
        self._channels = channels
        self._spk = _sc.default_speaker() if not id \
            else _sc.get_speaker(id)
        return

    def __call__(self, audio: Audio, blocking: bool = False):
        if self.nchannels != audio.nchannels:
            raise ValueError("The number of channels is incompatible.")
        self.frames = audio.nsamples
        if self.frames > self.nsamples:
            raise MemoryError("Requested playback time is greater than available space.")
        self.data[:self.frames] = audio[:]
        self._loop_wrapper(blocking)
        return

    @property
    def channels(self):
        """The device channels to output data to. Zero indexed."""
        return self._channels

    def _loop(self):
        with self._spk.player(self.samplerate, self.channels, self.blocksize) as p:
            self.running.set()
            while self.ridx < self.frames:
                p.play(self.read_next(self.blocksize//4))
                if self.finished.is_set():
                    break
        self.running.clear()
        self.finished.set()
        return

    def stop(self):
        if not self.finished.is_set():
            self.finished.set()
        self._thread.join()
        return

    def reset(self):
        self.ridx = 0
        return

    def get_playback(self, blocksize: int = None):
        return Audio(self.data[:self.frames].copy(), self.samplerate,
                     self.blocksize if not blocksize else blocksize)



# -*- coding: utf-8 -*-
"""
Utilities module.

Created on Tue May  5 00:31:42 2020

@author: João Vitor Gutkoski Paes
"""

from .colore import ColorStr, colorir, pinta_texto, pinta_fundo
from .freq import freq_to_band, fractional_octave_frequencies, normalize_frequencies, freqs_to_center_and_edges
from .maths import max_abs, rms, dB
from .logger import Logger, Now

__all__ = [
    # colore
    'ColorStr',
    'colorir',
    'pinta_texto',
    'pinta_fundo',

    # freq
    'freq_to_band',
    'fractional_octave_frequencies',
    'freqs_to_center_and_edges',
    'normalize_frequencies',

    # maths
    'max_abs',
    'rms',
    'dB',

    # logger
    'Logger',
    'Now']



# -*- coding: utf-8 -*-
"""
Ossom setup file
=================

@Author:
- João Vitor Gutkoski Paes, joao.paes@eac.ufsm.br

"""

from setuptools import setup

settings = {
    'name': 'OsSom',
    'version': '0.1.0a',
    'description': 'Audio IO tools for real time data visualization',
    'url': 'http://github.com/Chum4k3r/ossom',
    'author': 'João Vitor Gutkoski Paes',
    'author_email': 'joao.paes@eac.ufsm.br',
    'license': 'MIT',
    'install_requires': ['numpy', 'soundcard', 'numba'],
    'packages': ['ossom', 'ossom.utils'],
    'package_dir': {'utils': 'ossom'},
    'package_data': {'examples': ['examples/*.py']},
    'classifiers': [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    'python_requires': '>=3.8',
}

setup(**settings)




# -*- coding: utf-8 -*-
"""
Lacks documentation!

Created on Sat Jun 27 20:25:33 2020

@author: João Vitor Gutkoski Paes
"""

import numpy as np
import time
import multiprocessing as mp
from ossom import Recorder, Player, Configurations
from typing import Union


config = Configurations()


class Monitor(object):
    """Monitor class."""

    def __init__(self,
                 target: callable = lambda x: x,
                 samplerate: int = config.samplerate,
                 waittime: float = 1.,
                 args: tuple = (0,)):
        """
        Control a multiprocessing.Process to visualize data from recording or playing.

        Parameters
        ----------
        target : callable, optional
            DESCRIPTION. The default is None.
        args : tuple, optional
            DESCRIPTION. The default is None.

        Returns
        -------
        None.

        """
        self.target = target
        self.waitTime = waittime
        self.samplerate = samplerate
        self.readLen = int(np.ceil(waittime * samplerate))
        self.args = args
        return

    def __call__(self, strm: Union[Recorder, Player] = None, blocksize: int = None):
        """
        Configure the monitor and starts de process.

        Parameters
        ----------
        strm : Union[Recorder, Player], optional
            DESCRIPTION. The default is None.
        buffersize : int, optional
            DESCRIPTION. The default is None.

        Returns
        -------
        None.

        """
        self.running = strm.running
        self.finished = strm.finished
        self._buffer = strm.get_buffer(blocksize=self.readLen)
        # assert self.buffer.data is strm.data
        self._process = mp.Process(target=self._loop)
        return

    def setup(self):
        """Any setup step needed to the end monitoring object. Must be overriden on subclasses."""
        pass

    def tear_down(self):
        """Any destroying step needed to finish the end monitoring object. Must be overriden on subclasses."""
        pass

    def _loop(self):
        """
        Actual monitoring loop.

        Parameters
        ----------
        buffer : Audio
            DESCRIPTION.

        Returns
        -------
        None.

        """
        self.setup()
        self.running.wait()
        time.sleep(0.25)
        self.nextTime = time.time() + self.waitTime
        while self.running.is_set():
            sleepTime = self.nextTime - time.time()
            if sleepTime > 0.:
                time.sleep(sleepTime)
            self.target(self._buffer.read_next(self.readLen), *self.args)
            self.nextTime = time.time() + self.waitTime
            if self.finished.is_set():
                break
        self.tear_down()
        return

    def start(self):
        """Start the parallel process."""
        self._process.start()
        return

    def wait(self):
        """Finish the parallel process."""
        self._process.join()
        self._process.close()
        # self._buffer.close()
        return



# -*- coding: utf-8 -*-
"""
These classes are designed to provide objects that will handle reading and writing of audio data as numpy arrays.

They are two: `Audio`, which is a read-only object, and `AudioBuffer`, that is a subclass
of `Audio` and `multiprocessing.shared_memory.SharedMemory`, thus providing a cross-processes data read and write functionality.

Created on Fri May 29 16:07:04 2020.

@author: João Vitor Gutkoski Paes
"""

import numpy as np
import multiprocessing as mp
from multiprocessing import shared_memory as sm
from ossom import Configurations


config = Configurations()


class Audio(object):
    """Audio data interface."""

    def __init__(self, data: np.ndarray, samplerate: int,
                 blocksize: int = config.blocksize) -> None:
        """
        Audio objects are a representation of a waveform and a sample rate.

        They hold the basic information to play sound.

        Parameters
        ----------
        data : np.ndarray
            An array containing audio data, or a buffer to be filled with audio.
        samplerate : int
            Audio sample rate, or how many data represent one second of data.
        blocksize : int, optional
            Amount of samples to read on each call to `next`.
            The default is config.blocksize.

        Returns
        -------
        None

        """
        self._samplerate = int(samplerate)
        self._blocksize = blocksize
        self._data = data.reshape((-1, 1)) if data.ndim < 2 else data
        self._ridx = int()
        return

    def __getitem__(self, key):
        """Route Audio getitem to numpy.ndarray getitem."""
        return np.ndarray.__getitem__(self.data, key)

    def __iter__(self):
        """Iterate method."""
        return self

    def __next__(self) -> np.ndarray:
        """Return data from `ridx` to `ridx` + `blocksize`."""
        return self.read_next(self.blocksize)

    def read_next(self, blocksize: int) -> np.ndarray:
        """
        Read data from `ridx` to `ridx` + `blocksize`.

        Parameters
        ----------
        blocksize : int
            The amount of data to read.

        Raises
        ------
        StopIteration
            Unavailable to read more data than `nsamples`.

        Returns
        -------
        data : np.ndarray
            A numpy array with `blocksize` rows and `nchannels` columns.

        """
        if self.ridx > self.nsamples:
            raise StopIteration
        data = self.data[self.ridx:self.ridx+blocksize]
        self.ridx += blocksize
        return data

    @property
    def ridx(self) -> int:
        """Read data index."""
        return self._ridx

    @ridx.setter
    def ridx(self, n):
        self._ridx = n
        return

    @property
    def blocksize(self) -> int:
        """Amount of samples read on each call to `next()`."""
        return self._blocksize

    @property
    def samplerate(self) -> int:
        """Sample rate of the audio."""
        return self._samplerate

    @property
    def data(self):
        """Audio data as a numpy.ndarray."""
        return self._data

    @property
    def nsamples(self) -> int:
        """Total number of data."""
        return self.data.shape[0]

    @property
    def nchannels(self) -> int:
        """Total number of channels."""
        return self.data.shape[1]

    @property
    def duration(self) -> float:
        """Total time duration."""
        return self.nsamples/self.samplerate

    @property
    def samplesize(self) -> int:
        """Size of one sample of audio."""
        return self.data.itemsize

    @property
    def dtype(self) -> np.dtype:
        """Type of the data."""
        return self.data.dtype

    @property
    def bytesize(self) -> int:
        """Size, in bytes, of whole array. Same as `samplesize * nsamples * nchannels`."""
        return self.data.nbytes


class AudioBuffer(Audio, sm.SharedMemory):
    """Audio data in a shared memory buffer."""

    def __init__(self, name: str, samplerate: int,
                 buffersize: int, nchannels: int,
                 blocksize: int, dtype: np.dtype = config.dtype) -> None:
        """
        Buffer object intended to read and write audio samples.

        Parameters
        ----------
        name : str
            The SharedMemory name, can be None to automatically generate one.
        samplerate : int
            Audio sampling rate.
        buffersize : int
            Total number of samples.
        nchannels : int
            Total number of channels.
        blocksize : int
            Amount of samples to read on each call to `next`
        dtype : np.dtype
            Sample data type. The default is config.dtype.

        Returns
        -------
        None.

        """
        sz = dtype.itemsize * buffersize * nchannels

        if name is not None:
            try:
                sm.SharedMemory.__init__(self, name)
            except FileNotFoundError:
                sm.SharedMemory.__init__(self, name, create=True, size=sz)
        else:
            sm.SharedMemory.__init__(self, create=True, size=sz)
        buffer = np.ndarray((buffersize, nchannels),
                            dtype=dtype, buffer=self.buf)
        Audio.__init__(self, buffer, samplerate, buffersize)
        self._widx = mp.Value('i', int())
        self._full = mp.Event()
        self._full.clear()
        return

    def __del__(self):
        """Guarantee that SharedMemory calls close and unlink."""
        self.close()
        self.unlink()
        return

    @property
    def widx(self) -> int:
        """Write data index."""
        return self._widx.value

    @widx.setter
    def widx(self, n):
        self._widx.value = n
        if self._widx.value >= self.nsamples:
            self._full.set()
        return

    @property
    def ready2read(self) -> int or None:
        """
        How many samples are ready to read.

        If the write index is smaller than read index returns None

        Returns
        -------
        int
            Amount of samples available to read.

        """
        dif = self.widx - self.ridx
        return dif if dif > 0 else 0

    @property
    def is_full(self) -> bool:
        """Check if ringbuffer is full or not."""
        return self._full.is_set()

    def clear(self):
        """Set all data to zero."""
        self.data[:] = 0
        return

    def get_audio(self, blocksize: int = None, copy: bool = False) -> Audio:
        """
        `Audio` object that points to shared memory buffer, or a copy of it.

        Parameters
        ----------
        blocksize : int, optional
            The read block size. The default is None.
        copy : bool, optional
            If set to `True`, the `Audio` is just a copy of the buffer. The default is False.

        Returns
        -------
        Audio
            The buffer as a read-only Audio object.

        """
        return Audio(self.data if not copy else self.data.copy(),
                     self.samplerate, blocksize=blocksize if blocksize else self.blocksize)

    def write_next(self, data: np.ndarray) -> int or None:
        """
        Write data to buffer.

        If `widx` gets equal to `nsamples` the buffer is set to full.
        Checking can be made using the `is_full` property.

        Parameters
        ----------
        data : np.ndarray
            Samples to write on buffer.

        Returns
        -------
        int
            Amount of written samples.

        """
        wdata, wsz = self._write_check(data)
        self._data[self.widx:(self.widx + wsz)] = wdata[:]
        self.widx += wsz
        return

    def _write_check(self, data: np.ndarray):
        left = self.nsamples - self.widx
        return (data[:left], left) if data.shape[0] > left else (data, data.shape[0])
